/*
 * $Id: Expression.as,v 1.1.1.1 2006/07/17 03:27:32 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.util.bulletml {
  /**
   * Calculate four arithmetic operations with RPN.
   */
  public class Expression {
    private static const PLUS:int = -1;
    private static const MINUS:int = -2;
    private static const MULTIPLE:int = -3;
    private static const DIVISION:int = -4;
    private static const MODULO:int = -5;
    private static const RANK:int = -6;
    private static const RAND:int = -7;
    private static const VARIABLE:int = -8;
    private static var parsedExpressions:Array = new Array;
    
    public static function parse(s:String):Expression {
      if (parsedExpressions[s] != null)
        return parsedExpressions[s];
      var ep:Expression = new Expression;
      ep.parse(s);
      parsedExpressions[s] = ep;
      return ep;
    }

    private var str:String;
    private var operator:Array;
    private var value:Array;

    public function parse(s:String):void {
      str = s;
      operator = new Array;
      value = new Array;
      removeWhiteSpace();
      parseToRPN(0, str.length);
    }

    public function calc(params:Array, bullet:Bullet):Number {
      var stack:Array = new Array;
      for each(var op:int in operator) {
        if (op >= 0) {
          stack.push(value[op]);
        } else if (op <= VARIABLE) {
          stack.push(params[VARIABLE - op]);
        } else {
          switch(op) {
          case RAND:
            stack.push(bullet.getRand());
            break;
          case RANK:
            stack.push(bullet.getRank());
            break;
          case PLUS:
          case MINUS:
          case MULTIPLE:
          case DIVISION:
          case MODULO:
            var n1:Number = stack.pop();
            var n2:Number = stack.pop();
            stack.push(calcOperator(op, n2, n1));
            break;
          default:
            throw new Error("illegal operator: " + op);
          }
        }
      }
      return stack[0];
    }

    private function removeWhiteSpace():void {  
      var cs:String = new String;
      var skip:Boolean = false;
      var depth:int = 0;
      for (var i:int = 0; i < str.length; i++) {
        switch (str.charAt(i)) {
        case ' ':
        case '\n':
        case '\r':
          skip = true;
          break;
        case ')':
          depth--;
          if ( depth < 0 )
            throw new Error("bracket not match");
          break;
        case '(':
          depth++;
          break;
        }
        if ( skip )
          skip = false;
        else
          cs += str.charAt(i);
      }
      if (depth != 0)
        throw new Error("bracket not match");
      str = cs;
    }

    private function parseToRPN(stIdx:int, edIdx:int):void {
      var op0:int = -1;
      var op1:int = -1;
      for (var i:int = edIdx - 1; i >= stIdx; i--) {
        var c:String = str.charAt(i);
        if (c == ')') {
          do {
            i--;
          } while (str.charAt(i) != '(');
        } else if (op0 < 0 && (c == '*' || c == '/' || c == '%')) {
          op0 = i;
        } else if (c == '+' || c == '-') {
          op1 = i;
          break;
        }
      }
      if (op0 < 0 && op1 < 0 && str.charAt(stIdx) == '(' && str.charAt(edIdx-1) == ')') {
    	  parseToRPN(stIdx + 1, edIdx - 1);
    	  return;
      }
      if (op1 < 0) {
        if (op0 < 0) {
          parseFloatValue(stIdx, edIdx - stIdx);
        } else {
          parseToRPN(stIdx, op0);
          parseToRPN(op0 + 1, edIdx);
          switch(str.charAt(op0)) {
          case '*':
            pushOperator(MULTIPLE);
            break;
          case '/':
            pushOperator(DIVISION);
            break;
          case '%':
            pushOperator(MODULO);
            break;
          default:
            throw new Error("unknown operator: " + str.charAt(op0));
          }
        }
      } else {
        if (op1 == stIdx) {
          switch (str.charAt(op1)) {
          case '-':
            parseToRPN(stIdx + 1, edIdx);
            pushNumber(-1);
            pushOperator(MULTIPLE);
            break;
          case '+':
            parseToRPN(stIdx + 1, edIdx);
            break;
          default:
            throw new Error("unknown unary operator: " + str.charAt(op1));
          }
        } else {
          parseToRPN(stIdx, op1);
          parseToRPN(op1 + 1, edIdx);
          switch(str.charAt(op1)) {
          case '+':
            pushOperator(PLUS);
            break;
          case '-':
            pushOperator(MINUS);
            break;
          default:
            throw new Error("unknown operator: " + str.charAt(op1));
          }
        }
      }
    }

    private function parseFloatValue(stIdx:int, lgt:int):void {
      if (str.charAt(stIdx) == '$') {
        var label:String = str.substr(stIdx + 1, lgt - 1);
        if (label == "rand") {
          pushVariable(RAND);
        } else if (label == "rank") {
          pushVariable(RANK);
        } else {
          var nidx:Number = parseInt(label);
          if (isNaN(nidx))
            throw new Error("illegal variable: $" + label);
          var idx:int = nidx;
          idx--;
          pushVariable(VARIABLE - idx);
        }
      } else {
        var v:Number = Number(str.substr(stIdx, lgt));
        if (isNaN(v))
          throw new Error ("illegal number: " + str.substr(stIdx, lgt));
        pushNumber(v);
      }
    }

    private function pushOperator(op:int):void {
      var idx:int = operator.length;
      if (operator[idx - 2] >= 0 && operator[idx - 1] >= 0) {
        var i1:int = operator.pop();
        var i2:int = operator.pop();
        var n1:Number = value[i1];
        var n2:Number = value[i2];
        if (i1 > i2) {
          operator.splice(i1, 1);
          operator.splice(i2, 1);
        } else {
          operator.splice(i2, 1);
          operator.splice(i1, 1);
        }
        var ni:int = value.push(calcOperator(op, n2, n1)) - 1;
        operator.push(ni);
      } else {
        operator.push(op);
      }
    }

    private function pushVariable(vi:int):void {
      operator.push(vi);
    }

    private function pushNumber(n:Number):void {
      var ni:int = value.push(n) - 1;
      operator.push(ni);
    }

    private function calcOperator(op:int, n1:Number, n2:Number):Number {
      switch ( op ) {
      case PLUS:
        return n1 + n2;
      case MINUS:
        return n1 - n2;
      case MULTIPLE:
        return n1 * n2;
      case DIVISION:
        return n1 / n2;
      case MODULO:
        return n1 % n2;
      default:
        throw new Error("illegal operator: " + op);
      }
    }
  }
}
